# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from collections.abc import Callable
from datetime import timedelta
from itertools import chain
from math import floor, log10

import six

from more_itertools import collapse, partition

from dbnd._core.configuration.environ_config import in_airflow_tracking_mode
from dbnd._core.constants import SystemTaskName, TaskRunState
from dbnd._core.current import is_verbose
from dbnd._core.tracking.backends import TrackingStore
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.utils.basics.text_banner import safe_tabulate
from dbnd._core.utils.timezone import utcnow
from dbnd._vendor.ascii_graph import Pyasciigraph


try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

if typing.TYPE_CHECKING:
    from typing import Any, Dict, Iterable, List, Tuple

    from dbnd._core.task_run.task_run import TaskRun
    from targets.value_meta import ValueMeta

logger = logging.getLogger(__name__)


def is_describe_stat(item):
    name, _ = item
    return name in ["min_value", "quartile_1", "quartile_2", "quartile_3", "max_value"]


def two_columns_table(stats, split_by):
    # type: (Dict, Callable[Tuple[Any,Any], bool]) -> str
    """
    Split the items by a filter function "split_by" and merge
    Example:
        two_columns_table({1: "a", 2: "b", 3: "c", 4: "d"}, is_odd)
        [ 2 , b , 1 , a ]
        [ 3 , d , 3 , c ]
    """
    splatted = partition(split_by, stats.items())
    flatten_merged_rows = map(collapse, zip_longest(*splatted))
    return safe_tabulate(flatten_merged_rows, headers=())


class ConsoleStore(TrackingStore):
    def __init__(self, *args, **kwargs):
        super(ConsoleStore, self).__init__(*args, **kwargs)
        self.max_log_value_len = 50
        self.verbose = is_verbose()
        self.ascii_graph = Pyasciigraph()
        self._is_in_airflow_tracking_mode = in_airflow_tracking_mode()

    def init_run(self, run):
        if run.is_orchestration:
            logger.info(
                run.describe.run_banner(
                    "Running Databand v%s" % run.context.task_run_env.databand_version,
                    color="cyan",
                    show_run_info=True,
                )
            )

            if run.context.name == "interactive":
                from dbnd._core.tools import ipython

                ipython.show_run_url(run.run_url)

    def set_task_reused(self, task_run):
        task = task_run.task
        logger.info(
            task.ctrl.visualiser.banner(
                "Task %s has been completed already!" % task.task_id,
                "magenta",
                task_run=task_run,
            )
        )

    def set_task_run_state(self, task_run, state, error=None, timestamp=None):
        super(ConsoleStore, self).set_task_run_state(task_run=task_run, state=state)

        task = task_run.task

        if self._is_in_airflow_tracking_mode:
            if state == TaskRunState.RUNNING:
                logger.info(
                    "Tracking %s task at %s", task.task_name, task_run.task_tracker_url
                )
            return

        # optimize, don't print success banner for fast running tasks
        start_time = task_run.start_time or utcnow()
        quick_task = task_run.finished_time and (
            task_run.finished_time - start_time
        ) < timedelta(seconds=5)

        show_simple_log = not self.verbose and (
            task_run.task.task_is_system or quick_task
        )

        level = logging.INFO
        color = "cyan"
        task_friendly_id = task_run.task_af_id
        task_id_str = "%s(%s)" % (task.task_id, task_friendly_id)
        if state in [TaskRunState.RUNNING, TaskRunState.QUEUED]:
            task_msg = "Running task %s" % task_id_str

        elif state == TaskRunState.SUCCESS:
            task_msg = "Task %s has been completed!" % task_id_str
            color = "green"

        elif state == TaskRunState.FAILED:
            task_msg = "Task %s has failed!" % task_id_str
            color = "red"
            level = logging.WARNING
            if task_run.task.task_name in SystemTaskName.driver_and_submitter:
                show_simple_log = True

        elif state == TaskRunState.CANCELLED:
            task_msg = "Task %s has been canceled!" % task_id_str
            color = "red"
            level = logging.WARNING
            show_simple_log = True

        else:
            task_msg = "Task %s moved to %s state" % (task_friendly_id, state)

        if show_simple_log:
            logger.log(level, task_msg)
            return

        try:
            logger.log(
                level, task.ctrl.visualiser.banner(task_msg, color, task_run=task_run)
            )
        except Exception as ex:
            logger.log(level, "%s , failed to create banner: %s" % (task_msg, ex))

    def add_task_runs(self, run, task_runs):
        pass

    def save_external_links(self, task_run, external_links_dict):
        task = task_run.task
        attempt_uid = task_run.task_run_attempt_uid

        logger.info(
            "%s %s has registered URLs to external resources: %s",
            task,
            attempt_uid,
            ", ".join("%s: %s" % (k, v) for k, v in six.iteritems(external_links_dict)),
        )

    def is_ready(self):
        return True

    def log_histograms(self, task_run, key, value_meta, timestamp):
        # type: (TaskRun, str, ValueMeta, datetime) -> None

        for graph in self.get_console_histograms(key, value_meta):
            for line in graph:
                # we add one line at a time using the logger to keep the output aligned together
                logger.info(line)

        if not self.verbose:
            return
        if value_meta.histogram_system_metrics:
            for key, value in value_meta.histogram_system_metrics.items():
                if isinstance(value, float):
                    value = round(value, 3)
                logger.info("%s: %s", key, value)

    def get_console_histograms(self, key, value_meta):
        # type: (str, ValueMeta) -> Iterable[Iterable[str]]
        """
        Iterate over histograms and for each its build a all the lines to print to the console
        """
        for column_name, hist_values in value_meta.histograms.items():
            console_graph = self.ascii_graph.graph(
                label=f"Histogram logged: {key}.{column_name}",
                data=list(zip(*reversed(hist_values))),
            )

            # not will exists, but when it does - we need to add the stats as a table to the console
            if value_meta.columns_stats:
                column_stats = value_meta.get_column_stats_by_col_name(column_name)
                if column_stats:
                    stats_table = two_columns_table(
                        column_stats.as_dict(), is_describe_stat
                    )
                    # channing the printable graph and status table
                    console_graph = chain(console_graph, stats_table.splitlines())

            # separation line between one histogram and the other
            console_graph = chain(console_graph, [""])
            yield console_graph

    @staticmethod
    def _format_value(value):
        """
        Convert value to string.
        Additionally forces float values to NOT use scientific notation.
        """
        if isinstance(value, float):
            value_width = 1 if value == 0.0 else abs(floor(log10(abs(value))))
            # default python precision threshold to switch to scientific representation of floats
            if value > 1 or value_width < 4:
                return str(value)
            return "{:{width}.{prec}f}".format(
                value, width=value_width, prec=value_width + 4
            )
        return str(value)

    def log_metrics(self, task_run, metrics):
        # type: (TaskRun, List[Metric]) -> None
        if len(metrics) == 1:
            m = metrics[0]
            logger.info(
                "Task {name} metric: {key}={value}".format(
                    name=task_run.task_af_id,
                    key=m.key,
                    value=self._format_value(m.value),
                )
            )
        else:
            metrics_str = ", ".join(
                ["{}={}".format(m.key, self._format_value(m.value)) for m in metrics]
            )
            logger.info("Task %s metrics: %s", task_run.task_af_id, metrics_str)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import re
import typing

from dbnd._core.errors import DatabandError, friendly_error
from dbnd._core.task.task import Task
from dbnd._core.task.task_mixin import _TaskCtrlMixin
from dbnd._core.task_ctrl.task_ctrl import TaskSubCtrl
from dbnd._core.utils.task_utils import _try_get_task_from_airflow_op, to_tasks
from dbnd._core.utils.traversing import flatten


if typing.TYPE_CHECKING:
    from typing import Collection, Set

logger = logging.getLogger(__name__)


def _task_list(task_or_task_list):
    try:
        task_list = list(task_or_task_list)
    except TypeError:
        task_list = [task_or_task_list]

    from targets import Target

    task_list_output = []
    for t in task_list:
        if isinstance(t, Target):
            t = t.task
        airflow_task = _try_get_task_from_airflow_op(t)
        if airflow_task:
            t = airflow_task
        if not (isinstance(t, _TaskCtrlMixin)):
            raise DatabandError(
                "Relationships can only be set between "
                "Databand Tasks; received {}".format(t.__class__.__name__)
            )
        task_list_output.append(t)
    return task_list_output


class _TaskDagNode(TaskSubCtrl):
    def __init__(self, task):
        super(_TaskDagNode, self).__init__(task)

        self._upstream_tasks = set()
        self._downstream_tasks = set()

    def initialize_dag_node(self):
        # connect to all required tasks
        upstream = flatten(to_tasks(self.ctrl.task_inputs))
        upstream = list(filter(None, upstream))

        # take care of orphant tasks
        for child in self.task.descendants.get_children():
            if not child.task_dag.downstream:
                # it means there is no other tasks that are waiting for this task
                # -> we add it to task upstream
                # other options would be child is required by Parent task,
                # but it's not added yet ( anyway we create unique list)
                upstream.append(child)

        upstream = set(upstream)
        for upstream_task in upstream:
            self.set_upstream(upstream_task)

    def set_global_upstream(self, task_or_task_list):
        # don't set to itself for now, leafs are good enough
        # self.set_upstream(task_or_task_list=task_or_task_list)

        from dbnd.tasks import DataSourceTask

        children = {
            c
            for c in self.task.descendants.get_children()
            if not isinstance(c, DataSourceTask)
        }
        # do it only for leafs
        if not children:
            self.set_upstream(task_or_task_list=task_or_task_list)
        # or find the most "young child" (most upstream) and run it for him only
        # it's an ugly hack :)

        # let find all upstream tasks and intersect it with other children list.
        # the one with zero intersection is root in upstream!

        for c in children:
            upstreams = c.ctrl.task_dag.subdag_tasks()
            if not children & upstreams:
                c.set_global_upstream(task_or_task_list=task_or_task_list)

    def _task_id_to_tasks(self, task_ids):  # type: (Collection[str])->Set[Task]
        return {self.get_task_by_task_id(task_id) for task_id in task_ids}

    @property
    def upstream_task_ids(self):  # type: ()->Set[str]
        """list of tasks directly upstream"""
        return self._upstream_tasks

    @property
    def upstream(self):  # type: ()->Set[Task]
        """list of tasks directly upstream"""
        return self._task_id_to_tasks(self._upstream_tasks)

    @property
    def downstream_task_ids(self):  # type: ()->Set[str]
        """list of tasks directly downstream"""
        return self._downstream_tasks

    @property
    def downstream(self):  # type: ()->Set[Task]
        """list of tasks directly downstream"""
        return self._task_id_to_tasks(self._downstream_tasks)

    def set_upstream(self, task_or_task_list):
        self.set_relatives(task_or_task_list, upstream=True)

    def has_upstream(self, task):
        return task.task_id in self._upstream_tasks

    def set_downstream(self, task_or_task_list):
        self.set_relatives(task_or_task_list, upstream=False)

    def _direction(self, upstream):
        return self._upstream_tasks if upstream else self._downstream_tasks

    def subdag_tasks(self, should_run_only=False):
        return self._get_all_tasks(upstream=True, should_run_only=should_run_only)

    def _get_all_tasks(self, upstream=False, should_run_only=False):
        seen = set()
        result = []
        to_process = [self.task]
        # should be iterative, we don't like recursive as we can have huge nesting
        while to_process:
            current = to_process.pop(0)
            seen.add(current.task_id)
            if should_run_only and not current.ctrl.should_run():
                continue
            t_dag = current.ctrl.task_dag
            result.append(current)
            for t_connected_task_id in t_dag._direction(upstream):
                if t_connected_task_id in seen:
                    continue
                connected_task = self.get_task_by_task_id(t_connected_task_id)
                if not connected_task:
                    raise DatabandError(
                        "Can't resolve task %s by it's id" % t_connected_task_id
                    )
                to_process.append(connected_task)

        return set(result)

    def set_relatives(self, task_or_task_list, upstream=False):
        task_list = _task_list(task_or_task_list)
        for task in task_list:
            task.ctrl.task_dag._set_relatives(self.task, upstream=not upstream)
            self._set_relatives(task, upstream=upstream)

    def _set_relatives(self, task, upstream=False):
        connected = self._direction(upstream)
        if task.task_id in connected:
            connected_task = self.get_task_by_task_id(task.task_id)
            if id(task) == id(connected_task):
                logger.warning(
                    "Dependency {task} already registered for {current}"
                    "".format(task=task, current=self.task)
                )
            else:
                logger.debug("Re adding new implementation %s to %s", task, self.task)
        else:
            connected.add(task.task_id)

    def topological_sort(self):
        """
        Sorts tasks in topographical order, such that a task comes after any of its
        upstream dependencies.

        Heavily inspired by:
        http://blog.jupo.org/2012/04/06/topological-sorting-acyclic-directed-graphs/

        :return: list of tasks in topological order
        """

        tasks = self.subdag_tasks()
        return topological_sort(tasks)

    def select_by_task_names(self, tasks_regexes, tasks=None):
        tasks = tasks or self.subdag_tasks()
        selected = []
        for task_regex in tasks_regexes:
            for t in tasks:
                if re.findall(task_regex, t.task_id):
                    selected.append(t)
        if not selected:
            raise friendly_error.no_matching_tasks_in_pipeline(
                tasks=tasks, tasks_regexes=tasks_regexes
            )
        return selected


def topological_sort(tasks, root_task=None):
    # special case
    if len(tasks) == 0:
        return tuple()

    graph_unsorted = set(tasks)

    graph_sorted = []

    # Run until the unsorted graph is empty.
    while graph_unsorted:
        # Go through each of the node/edges pairs in the unsorted
        # graph. If a set of edges doesn't contain any nodes that
        # haven't been resolved, that is, that are still in the
        # unsorted graph, remove the pair from the unsorted graph,
        # and append it to the sorted graph. Note here that by using
        # using the items() method for iterating, a copy of the
        # unsorted graph is used, allowing us to modify the unsorted
        # graph as we move through it. We also keep a flag for
        # checking that that graph is acyclic, which is true if any
        # nodes are resolved during each pass through the graph. If
        # not, we need to bail out as the graph therefore can't be
        # sorted.
        acyclic = False
        for node in list(graph_unsorted):
            for edge in node.ctrl.task_dag.upstream:
                if edge in graph_unsorted:
                    break
            # no edges in upstream tasks
            else:
                # we found at least one node in graph
                acyclic = True
                graph_unsorted.remove(node)
                graph_sorted.append(node)

        if not acyclic:
            raise friendly_error.graph.cyclic_graph_detected(root_task, graph_unsorted)

    return tuple(graph_sorted)


def all_subdags(tasks):
    result = set()
    for sub_root in tasks:
        result = result.union(set(sub_root.task_dag.subdag_tasks()))
    return result



# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import os

import six


def generate_fernet_key():
    try:
        from cryptography.fernet import Fernet
    except ImportError:
        return ''
    else:
        return Fernet.generate_key().decode()


def expand_env_var(env_var):
    """
    Expands (potentially nested) env vars by repeatedly applying
    `expandvars` and `expanduser` until interpolation stops having
    any effect.
    """
    if not env_var:
        return env_var
    if not isinstance(env_var, six.string_types):
        # we don't support interpolation of non string values for now
        # sometime configparser can have complex structures, without this condition
        # this will make that objects strings
        return env_var
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(env_var))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated



# © Copyright Databand.ai, an IBM Company 2022

import logging

import six

from dbnd._core.current import try_get_current_task_run


logger = logging.getLogger(__name__)


def set_external_resource_urls(links):
    # type: (dict[str, str]) -> None
    """
    Will add the `links` as external resources, to the current running task.
    If there is no running task or the links are bad formatted it will fail without raising.

    @param links: map between the name of the resource and the link to the resource
    """
    if not isinstance(links, dict):
        logger.warning("Failed to add links as external resources: links is not a dict")
        return

    links = {item: value for item, value in six.iteritems(links) if value}
    if not links:
        logger.warning(
            "Failed to add links as external resources: links is empty dict or dict with None values "
        )
        return

    task_run = try_get_current_task_run()
    if task_run is None:
        logger.warning(
            "Failed to add links as external resources: There is no running task"
        )
        return

    try:
        task_run.set_external_resource_urls(links)
    except Exception as e:
        logger.warning(
            "Failed to add links as external resources: raised {e}".format(e=e)
        )



# © Copyright Databand.ai, an IBM Company 2022

import typing

from typing import List

from dbnd._core.run.run_ctrl import RunCtrl


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun
    from dbnd._core.settings import EngineConfig
    from dbnd._core.task_run.task_run import TaskRun


class TaskExecutor(RunCtrl):
    def __init__(self, run, task_executor_type, host_engine, target_engine, task_runs):
        # type: (DatabandRun, str, EngineConfig, EngineConfig,  List[TaskRun])->None
        super(TaskExecutor, self).__init__(run=run)

        self.task_executor_type = task_executor_type
        self.host_engine = host_engine
        self.target_engine = target_engine
        self.task_runs = task_runs

    def do_run(self):
        raise NotImplementedError()



# © Copyright Databand.ai, an IBM Company 2022

import json
import logging

from dbnd._core.constants import AlertDefOperator
from dbnd._core.current import get_databand_context
from dbnd.api.jobs import is_job_exists
from dbnd.api.scheduler import is_scheduled_job_exists
from dbnd.api.shared_schemas.alerts_def_schema import AlertDefsSchema


logger = logging.getLogger(__name__)


def run_if_job_exists(func):
    def inner(job_name, *args, **kwargs):
        if is_job_exists(job_name) or is_scheduled_job_exists(job_name):
            return func(job_name, *args, **kwargs)
        else:
            raise LookupError("Job named '%s' is not found" % job_name)

    return inner


def _post_alert(client, alert):
    alert_uid = client.api_request(endpoint="alert_defs", data=alert, method="POST")
    return alert_uid


def create_alert(
    job_name,
    job_id,
    task_name,
    task_repr,
    uid,
    alert_class,
    severity,
    operator,
    value,
    user_metric,
    is_str_value=False,
    **optional_fields,
):
    alert = {
        "type": alert_class,
        "job_name": job_name,
        "job_id": job_id,
        "task_name": task_name,
        "task_repr": task_repr,
        "uid": uid,
        "operator": operator,
        "value": value,
        "severity": severity,
        "user_metric": user_metric,
        "is_str_value": is_str_value,
    }
    if operator == AlertDefOperator.ANOMALY.value:
        alert["ml_alert"] = {
            "look_back": optional_fields["look_back"],
            "sensitivity": optional_fields["sensitivity"],
        }
    if operator == AlertDefOperator.RANGE.value:
        alert["operator"] = "RangeAlert"
        alert["value"] = json.dumps(
            {"baseline": optional_fields["baseline"], "range": optional_fields["range"]}
        )
    alert = {key: value for key, value in alert.items() if value is not None}
    alert_def_uid = _post_alert(get_databand_context().databand_api_client, alert)
    return alert_def_uid


@run_if_job_exists
def list_job_alerts(job_name):
    return get_alerts_filtered(job_name=job_name)


def _get_alerts(client, query_api_args):
    endpoint = "?".join(["alert_defs", query_api_args])
    schema = AlertDefsSchema(strict=False)
    response = client.api_request(endpoint=endpoint, data="", method="GET")
    return schema.load(data=response["data"], many=True).data


def get_alerts_filtered(alert_def_uid=None, job_name=None, custom_name=None):
    """get alerts by filters"""
    params = {"uids": alert_def_uid, "job_name": job_name, "custom_name": custom_name}
    query_params = "&".join(
        f"{key}={value}" for key, value in params.items() if value is not None
    )
    return _get_alerts(get_databand_context().databand_api_client, query_params)


def delete_alerts(uids):
    """delete alerts by uids"""
    get_databand_context().databand_api_client.api_request(
        endpoint="alert_defs/delete", data=uids, method="POST"
    )



# © Copyright Databand.ai, an IBM Company 2022

from typing import List, Optional


def get_project_name_from_airflow_tags(tags):
    # type: (Optional[List[str]]) -> Optional[str]
    if not tags:
        return None

    project_tag_prefix = "project:"
    for tag in sorted(tags):
        if tag.startswith(project_tag_prefix):
            return tag.replace(project_tag_prefix, "").strip() or None



# © Copyright Databand.ai, an IBM Company 2022

import logging
import sys

from dbnd._core.utils.basics.environ_utils import environ_enabled


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


logger = logging.getLogger(__name__)

ENV_DBND__VERBOSE = "DBND__VERBOSE"  # VERBOSE
_VERBOSE = environ_enabled(ENV_DBND__VERBOSE)


def is_verbose():
    return _VERBOSE


def set_verbose(verbose: bool = True):
    global _VERBOSE
    _VERBOSE = verbose


def dbnd_log_debug(msg, *args, **kwargs):
    try:
        if is_verbose():
            logger.info(msg, *args, **kwargs)
        else:
            logger.debug(msg, *args, **kwargs)
    except:
        eprint("Failed to print dbnd info message")


def dbnd_log_info_error(msg, *args, **kwargs):
    """we show exception only in verbose mode"""
    try:
        if is_verbose():
            logger.exception(msg, *args, **kwargs)
        else:
            logger.info(msg, *args, **kwargs)
    except Exception:
        eprint("Failed to print dbnd error message")


def dbnd_log_exception(msg, *args, **kwargs):
    """we show exception only in verbose mode"""
    try:
        if is_verbose():
            logger.exception(msg, *args, **kwargs)
        else:
            logger.info(msg, *args, **kwargs)
    except Exception:
        eprint("Failed to print dbnd error message")


def dbnd_log_tracking(msg, *args, **kwargs):
    if _VERBOSE:
        eprint("DBND TRACKING: %s" % msg)


def dbnd_log_init_msg(msg):
    if _VERBOSE:
        eprint("DBND INIT: %s" % msg)



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import hashlib
import uuid

import pytz
import six

from dbnd._core.configuration.environ_config import get_dbnd_project_config
from dbnd._core.utils.basics.memoized import cached
from dbnd._vendor import pendulum


def source_md5(source_code):
    if source_code:
        try:
            return hashlib.md5(source_code.encode("utf-8")).hexdigest()  # nosec B324
        except UnicodeDecodeError:
            return hashlib.md5(source_code).hexdigest()  # nosec B324


def get_uuid():
    # TODO: obfuscate getnode() - mac address part
    return uuid.uuid1()


NAMESPACE_DBND = uuid.uuid5(uuid.NAMESPACE_DNS, "databand.ai")
NAMESPACE_DBND_JOB = uuid.uuid5(NAMESPACE_DBND, "job")
NAMESPACE_DBND_RUN = uuid.uuid5(NAMESPACE_DBND, "run")
NAMESPACE_DBND_TASK_DEF = uuid.uuid5(NAMESPACE_DBND, "task_definition")


def get_stable_uid(payload: str, namespace=uuid.NAMESPACE_DNS):
    return uuid.uuid5(namespace, payload)


def get_task_def_uid(dag_id, task_id, code_hash):
    return uuid.uuid5(
        NAMESPACE_DBND_TASK_DEF, "{}.{}.{}".format(dag_id, task_id, code_hash)
    )


def get_task_run_uid(run_uid, dag_id, task_id):
    return uuid.uuid5(run_uid, "{}.{}".format(dag_id, task_id))


def get_task_run_attempt_uid(run_uid, dag_id, task_id, try_number):
    return uuid.uuid5(run_uid, "{}.{}:{}".format(dag_id, task_id, try_number))


def get_task_run_attempt_uid_for_subrun(task_run):
    return uuid.uuid5(
        task_run.run.run_uid,
        "{}.{}:{}-subrun".format(
            task_run.run.dag_id, task_run.task_af_id, task_run.attempt_number
        ),
    )


def get_task_run_attempt_uid_by_task_run(task_run):
    is_sub_rub = get_dbnd_project_config().resubmit_run
    if is_sub_rub:
        return get_task_run_attempt_uid_for_subrun(task_run)

    return get_task_run_attempt_uid(
        task_run.run.run_uid,
        task_run.run.dag_id,
        task_run.task_af_id,
        task_run.attempt_number,
    )


def get_task_run_attempt_uid_from_af_ti(ti):
    airflow_instance_uid = get_airflow_instance_uid()
    run_uid = get_job_run_uid(
        airflow_instance_uid=airflow_instance_uid,
        dag_id=ti.dag_id,
        execution_date=ti.execution_date,
    )
    return get_task_run_attempt_uid(run_uid, ti.dag_id, ti.task_id, ti.try_number)


def get_job_run_uid(airflow_instance_uid, dag_id, execution_date):
    # TODO_CORE: change to source_instance_uid
    if isinstance(execution_date, six.string_types):
        execution_date = pendulum.parse(execution_date)
    if isinstance(execution_date, datetime.datetime):
        execution_date = execution_date.astimezone(pytz.utc).isoformat()
    if airflow_instance_uid is None:
        return uuid.uuid5(NAMESPACE_DBND_RUN, "{}:{}".format(dag_id, execution_date))
    else:
        return uuid.uuid5(
            NAMESPACE_DBND_RUN,
            "{}:{}:{}".format(airflow_instance_uid, dag_id, execution_date),
        )


def get_job_uid(airflow_server_info_uid, dag_id):
    # TODO_CORE: change to source_instance_uid

    if airflow_server_info_uid:
        return uuid.uuid5(
            NAMESPACE_DBND_JOB, "{}:{}".format(airflow_server_info_uid, dag_id)
        )
    else:
        return uuid.uuid5(NAMESPACE_DBND_JOB, dag_id)


@cached()
def get_airflow_instance_uid():
    """used to distinguish between jobs of different airflow instances"""
    import airflow

    db_url = airflow.settings.Session.bind.engine.url
    db_str = "{}:{}/{}".format(db_url.host, db_url.port, db_url.database)
    airflow_instance_uid = uuid.uuid5(uuid.NAMESPACE_URL, db_str)
    return str(airflow_instance_uid)


def get_dataset_op_uid(dataset_uid, task_run_attempt_id, operation_type):
    return uuid.uuid5(dataset_uid, "{}.{}".format(task_run_attempt_id, operation_type))



# © Copyright Databand.ai, an IBM Company 2022

import contextlib
import sys

from collections.abc import Callable

import six

from dbnd._vendor import cloudpickle


def qualname_func(func):
    return func.__qualname__


try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

try:
    import_errors = (ImportError, ModuleNotFoundError)
except Exception:
    # we are python2
    import_errors = (ImportError,)

try:
    from thread import get_ident
except ImportError:
    try:
        from _thread import get_ident
    except ImportError:
        from _dummy_thread import get_ident


def fix_sys_path_str():
    if six.PY2:
        # fix path from "non" str values, otherwise we fail on py2
        sys.path = [str(p) if type(p) != str else p for p in sys.path]


try:
    import urlparse
except ImportError:
    from urllib.parse import urlparse

__all__ = [
    "contextlib",
    "cloudpickle",
    "qualname_func",
    "import_errors",
    "Callable",
    "get_ident",
    "fix_sys_path_str",
    "urlparse",
    "StringIO",
]



# © Copyright Databand.ai, an IBM Company 2022

from typing import Any

from targets import Target
from targets.target_config import TargetConfig
from targets.values import (
    ObjectValueType,
    ValueType,
    get_value_type_of_obj,
    get_value_type_of_type,
)


_NOTHING = object()


class InMemoryTarget(Target):
    def __init__(self, obj=_NOTHING, path=None, value_type=None, **kwargs):
        # type: (Any, str, ValueType, **Any) -> None
        from targets.values import get_value_type_of_obj

        super(InMemoryTarget, self).__init__(**kwargs)
        self._obj = obj
        if self._obj is _NOTHING and not path:
            raise Exception("InMemoryTarget requires object or path")

        self.value_type = value_type or get_value_type_of_obj(self._obj)
        self.path = path or "memory://%s:%s" % (
            self.value_type,
            self.value_type.to_signature(self._obj),
        )
        self.config = TargetConfig()

    def exists(self):
        return self._obj is not _NOTHING

    def dump(self, value, value_type=None, **kwargs):
        self._obj = value
        if value_type:
            self.value_type = get_value_type_of_type(value_type)
        else:
            self.value_type = get_value_type_of_obj(value, ObjectValueType())

    def load(self, **kwargs):
        if not self.exists():
            raise ValueError("Value not assigned")
        return self._obj

    def __repr__(self):
        if isinstance(self._obj, Target):
            return str(self._obj)

        return self.path

    def __str__(self):
        return repr(self)



# -*- coding: utf-8 -*-
#
# Copyright 2012-2015 Spotify AB
# Modifications copyright (C) 2018 databand.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from dbnd._vendor.snippets.edit_distance import get_editdistance


def module_parents(module_name):
    """
    >>> list(module_parents('a.b'))
    ['a.b', 'a', '']
    """
    spl = module_name.split(".")
    for i in range(len(spl), 0, -1):
        yield ".".join(spl[0:i])
    if module_name:
        yield ""


def get_best_candidate(task_name, known_task_names, max_dist=5):
    weighted_tasks = [
        (get_editdistance(str(task_name), known_task_name), known_task_name)
        for known_task_name in known_task_names
    ]
    ordered_tasks = sorted(weighted_tasks, key=lambda pair: pair[0])
    return [
        task for (dist, task) in ordered_tasks if dist <= max_dist and dist < len(task)
    ]



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging
import typing

import dbnd

from dbnd._core.constants import (
    DbndDatasetOperationType,
    DbndTargetOperationStatus,
    DbndTargetOperationType,
)
from dbnd._core.tracking.backends.abstract_tracking_store import TrackingStore
from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.tracking.tracking_info_convertor import TrackingInfoBuilder
from dbnd._core.utils import json_utils
from dbnd._core.utils.timezone import utcnow
from dbnd.api.tracking_api import (
    LogDatasetArgs,
    LogTargetArgs,
    TaskRunAttemptUpdateArgs,
    add_task_runs_schema,
    airflow_task_infos_schema,
    heartbeat_schema,
    init_run_schema,
    log_artifact_schema,
    log_datasets_schema,
    log_dbt_metadata_schema,
    log_metrics_schema,
    log_targets_schema,
    save_external_links_schema,
    save_task_run_log_schema,
    scheduled_job_args_schema,
    set_run_state_schema,
    set_task_run_reused_schema,
    set_unfinished_tasks_state_schema,
    update_task_run_attempts_schema,
)
from targets import Target
from targets.value_meta import ValueMeta


if typing.TYPE_CHECKING:
    from typing import Any, Iterable, List, Optional, Union
    from uuid import UUID

    from dbnd._core.constants import TaskRunState
    from dbnd._core.task_run.task_run import TaskRun
    from dbnd._core.task_run.task_run_error import TaskRunError

logger = logging.getLogger(__name__)


class TrackingStoreThroughChannel(TrackingStore):
    """Track data to Tracking API"""

    def __init__(self, channel: TrackingChannel, *args, **kwargs):
        super(TrackingStoreThroughChannel, self).__init__(*args, **kwargs)
        self.channel = channel

    def init_scheduled_job(self, scheduled_job, update_existing):
        marsh = scheduled_job_args_schema.dump(
            dict(scheduled_job_args=scheduled_job, update_existing=update_existing)
        )
        resp = self.channel.init_scheduled_job(marsh.data)
        return resp

    def init_run(self, run):
        init_args = TrackingInfoBuilder(run).build_init_args()
        return self.init_run_from_args(init_args=init_args)

    def init_run_from_args(self, init_args):
        marsh = init_run_schema.dump(
            dict(init_args=init_args, version=dbnd.__version__)
        )
        resp = self.channel.init_run(marsh.data)
        return resp

    def add_task_runs(self, run, task_runs):
        task_runs_info = TrackingInfoBuilder(run).build_task_runs_info(
            task_runs=task_runs, dynamic_task_run_update=True
        )
        marsh = add_task_runs_schema.dump(
            dict(task_runs_info=task_runs_info, source=run.source)
        )
        resp = self.channel.add_task_runs(marsh.data)
        return resp

    def set_run_state(self, run, state, error=None, timestamp=None):
        marsh = set_run_state_schema.dump(
            dict(run_uid=run.run_uid, state=state, timestamp=timestamp)
        )
        resp = self.channel.set_run_state(marsh.data)
        return resp

    def set_task_reused(self, task_run):
        marsh = set_task_run_reused_schema.dump(
            dict(
                task_run_uid=task_run.task_run_uid,
                task_outputs_signature=task_run.task.task_outputs_signature_obj.signature,
            )
        )
        resp = self.channel.set_task_reused(marsh.data)
        return resp

    def set_task_run_state(self, task_run, state, error=None, timestamp=None):
        # type: (TaskRun, TaskRunState, TaskRunError, datetime.datetime) -> None
        marsh = update_task_run_attempts_schema.dump(
            dict(
                task_run_attempt_updates=[
                    TaskRunAttemptUpdateArgs(
                        task_run_uid=task_run.task_run_uid,
                        task_run_attempt_uid=task_run.task_run_attempt_uid,
                        state=state,
                        error=error.as_error_info() if error else None,
                        timestamp=timestamp or utcnow(),
                        source=task_run.run.source,
                    )
                ]
            )
        )
        resp = self.channel.update_task_run_attempts(marsh.data)
        return resp

    def set_task_run_states(self, task_runs):
        marsh = update_task_run_attempts_schema.dump(
            dict(
                task_run_attempt_updates=[
                    TaskRunAttemptUpdateArgs(
                        task_run_uid=task_run.task_run_uid,
                        task_run_attempt_uid=task_run.task_run_attempt_uid,
                        state=task_run.task_run_state,
                        timestamp=utcnow(),
                        source=task_run.run.source,
                    )
                    for task_run in task_runs
                ]
            )
        )
        resp = self.channel.update_task_run_attempts(marsh.data)
        return resp

    def set_unfinished_tasks_state(self, run_uid, state):
        marsh = set_unfinished_tasks_state_schema.dump(
            dict(run_uid=run_uid, state=state, timestamp=utcnow())
        )
        resp = self.channel.set_unfinished_tasks_state(marsh.data)
        return resp

    def update_task_run_attempts(self, task_run_attempt_updates):
        marsh = update_task_run_attempts_schema.dump(
            dict(task_run_attempt_updates=task_run_attempt_updates)
        )
        resp = self.channel.update_task_run_attempts(marsh.data)
        return resp

    def save_task_run_log(self, task_run, log_body, local_log_path=None):
        marsh = save_task_run_log_schema.dump(
            dict(
                task_run_attempt_uid=task_run.task_run_attempt_uid,
                log_body=log_body,
                local_log_path=local_log_path,
            )
        )
        resp = self.channel.save_task_run_log(marsh.data)
        return resp

    def save_external_links(self, task_run, external_links_dict):
        marsh = save_external_links_schema.dump(
            dict(
                task_run_attempt_uid=task_run.task_run_attempt_uid,
                external_links_dict=external_links_dict,
            )
        )
        resp = self.channel.save_external_links(marsh.data)
        return resp

    def log_dataset(
        self,
        task_run,  # type: TaskRun
        operation_path,  # type: Union[Target, str]
        data_meta,  # type: ValueMeta
        operation_type,  # type: DbndDatasetOperationType
        operation_status,  # type: DbndTargetOperationStatus
        operation_error,  # type: str
        operation_source=None,  # type: Optional[str]
        with_partition=None,  # type: Optional[bool]
    ):
        dataset_info = LogDatasetArgs(
            run_uid=task_run.run.run_uid,
            task_run_uid=task_run.task_run_uid,
            task_run_name=task_run.job_name,
            task_run_attempt_uid=task_run.task_run_attempt_uid,
            operation_path=str(operation_path),
            operation_type=operation_type,
            operation_status=operation_status,
            operation_error=operation_error,
            operation_source=operation_source,
            value_preview=data_meta.value_preview,
            columns_stats=data_meta.columns_stats,
            data_dimensions=data_meta.data_dimensions,
            data_schema=data_meta.data_schema,
            query=data_meta.query,
            with_partition=with_partition,
            timestamp=utcnow(),
            dataset_uri=None,
        )
        res = self.log_datasets(datasets_info=[dataset_info])
        return res

    def log_datasets(self, datasets_info):  # type: (List[LogDatasetArgs]) -> Any
        marsh = log_datasets_schema.dump(dict(datasets_info=datasets_info))
        resp = self.channel.log_datasets(marsh.data)
        return resp

    def log_dbt_metadata(self, dbt_run_metadata, task_run):
        marsh = log_dbt_metadata_schema.dump(
            dict(
                dbt_run_metadata=dbt_run_metadata,
                task_run_attempt_uid=task_run.task_run_attempt_uid,
            )
        )
        resp = self.channel.log_dbt_metadata(marsh.data)
        return resp

    def log_target(
        self,
        task_run,
        target,
        target_meta,  # type: ValueMeta
        operation_type,  # type: DbndTargetOperationType
        operation_status,  # type: DbndTargetOperationStatus
        param_name=None,  # type: Optional[str]
        task_def_uid=None,  # type: Optional[UUID]
    ):
        data_schema = (
            json_utils.dumps(target_meta.data_schema.as_dict())
            if target_meta.data_schema is not None
            else None
        )
        target_info = LogTargetArgs(
            run_uid=task_run.run.run_uid,
            task_run_uid=task_run.task_run_uid,
            task_run_name=task_run.job_name,
            task_run_attempt_uid=task_run.task_run_attempt_uid,
            task_def_uid=task_def_uid,
            param_name=param_name,
            target_path=str(target),
            operation_type=operation_type,
            operation_status=operation_status,
            value_preview=target_meta.value_preview,
            data_dimensions=target_meta.data_dimensions,
            data_schema=data_schema,
            data_hash=target_meta.data_hash,
        )
        res = self.log_targets(targets_info=[target_info])
        if target_meta.histograms:
            self.log_histograms(task_run, param_name, target_meta, utcnow())
        return res

    def log_targets(self, targets_info):  # type: (List[LogTargetArgs]) -> None
        marsh = log_targets_schema.dump(dict(targets_info=targets_info))
        resp = self.channel.log_targets(marsh.data)
        return resp

    def log_histograms(self, task_run, key, value_meta, timestamp):
        value_meta_metrics = value_meta.build_metrics_for_key(key)
        if value_meta_metrics["histograms"]:
            self.log_metrics(
                task_run=task_run, metrics=value_meta_metrics["histograms"]
            )

    def log_metrics(self, task_run, metrics):
        # type: (TaskRun, Iterable[Metric]) -> None
        metrics_info = [
            {
                "task_run_attempt_uid": task_run.task_run_attempt_uid,
                "metric": metric,
                "source": metric.source,
            }
            for metric in metrics
        ]
        marsh = log_metrics_schema.dump(dict(metrics_info=metrics_info))
        resp = self.channel.log_metrics(marsh.data)
        return resp

    def log_artifact(self, task_run, name, artifact, artifact_target):
        marsh = log_artifact_schema.dump(
            dict(
                task_run_attempt_uid=task_run.task_run_attempt_uid,
                name=name,
                path=artifact_target.path,
            )
        )
        resp = self.channel.log_artifact(marsh.data)
        return resp

    def heartbeat(self, run_uid):
        marsh = heartbeat_schema.dump(dict(run_uid=run_uid))
        resp = self.channel.heartbeat(marsh.data)
        return resp

    def save_airflow_task_infos(self, airflow_task_infos, source, base_url):
        marsh = airflow_task_infos_schema.dump(
            dict(
                airflow_task_infos=airflow_task_infos, source=source, base_url=base_url
            )
        )
        resp = self.channel.save_airflow_task_infos(marsh.data)
        return resp

    def flush(self):
        self.channel.flush()

    def is_ready(self):
        return self.channel.is_ready()

    def __str__(self):
        return "TrackingStoreThroughChannel with channel=%s" % (str(self.channel),)

    @staticmethod
    def build_with_disabled_channel(databand_ctx):
        from dbnd._core.tracking.backends.channels.tracking_disabled_channel import (
            DisabledTrackingChannel,
        )

        return TrackingStoreThroughChannel(channel=DisabledTrackingChannel())

    @staticmethod
    def build_with_console_debug_channel(databand_ctx):
        from dbnd._core.tracking.backends.channels.tracking_debug_channel import (
            ConsoleDebugTrackingChannel,
        )

        return TrackingStoreThroughChannel(channel=ConsoleDebugTrackingChannel())

    @staticmethod
    def build_with_web_channel(databand_ctx):
        from dbnd._core.tracking.backends.channels.tracking_web_channel import (
            TrackingWebChannel,
        )

        return TrackingStoreThroughChannel(
            channel=TrackingWebChannel(
                databand_api_client=databand_ctx.databand_api_client
            )
        )

    @staticmethod
    def build_with_async_web_channel(databand_ctx):
        from dbnd._core.tracking.backends.channels.tracking_async_web_channel import (
            TrackingAsyncWebChannel,
        )

        parameters = {
            "max_retries": databand_ctx.settings.core.max_tracking_store_retries,
            "remove_failed_store": databand_ctx.settings.core.remove_failed_store,
            "tracker_raise_on_error": databand_ctx.settings.core.tracker_raise_on_error,
            "is_verbose": databand_ctx.system_settings.verbose,
            "databand_api_client": databand_ctx.databand_api_client,
        }

        return TrackingStoreThroughChannel(
            channel=TrackingAsyncWebChannel(**parameters)
        )



# © Copyright Databand.ai, an IBM Company 2022

from contextlib import contextmanager

from dbnd._core.utils.basics.patch_properties import patched_properties


@contextmanager
def ready_for_pickle(obj, unpicklable_properties):
    """
    Remove unpicklable properties before dump obj and resume them after.
    This method could be called in dump method, to ensure unpicklable properties won't break dump.
    This method is a context-manager which can be called as below:

    .. code-block: python

        class MyObj(object):

            def _dump(self):
                with no_unpicklable_properties(self,("some_property",)):
                    pickle.dumps(self)

    """
    patch_properties = []
    for p in unpicklable_properties:
        if isinstance(p, tuple):
            property_name, property_value = p[0], p[1]
        else:
            property_name, property_value = p, "_non_pickable_property_"
        if hasattr(obj, property_name):
            patch_properties.append((property_name, property_value))

    with patched_properties(obj, patch_properties) as obj:
        yield obj



# © Copyright Databand.ai, an IBM Company 2022

import os
import typing

from datetime import date, datetime
from typing import Optional
from uuid import UUID

import attr

from dbnd._core.configuration.environ_config import (
    DBND_PARENT_TASK_RUN_ATTEMPT_UID,
    DBND_PARENT_TASK_RUN_UID,
    DBND_ROOT_RUN_TRACKER_URL,
    DBND_ROOT_RUN_UID,
    SCHEDULED_DAG_RUN_ID_ENV,
    SCHEDULED_DATE_ENV,
    SCHEDULED_JOB_UID_ENV,
)
from dbnd._core.constants import RunState, _DbndDataClass
from dbnd._core.current import try_get_current_task_run, try_get_databand_run
from dbnd._core.tracking.dbnd_spark_init import get_value_from_spark_env
from dbnd._core.utils import timezone


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun


@attr.s
class RunInfo(_DbndDataClass):
    run_uid = attr.ib()  # type: UUID

    job_name = attr.ib()  # type: str
    user = attr.ib()  # type: str

    name = attr.ib()  # type: str
    description = attr.ib()  # type: str

    state = attr.ib()  # type: RunState
    start_time = attr.ib()  # type: datetime
    end_time = attr.ib()  # type: Optional[datetime]

    # deprecate
    dag_id = attr.ib()  # type: str
    execution_date = attr.ib()  # type: datetime
    cmd_name = attr.ib()  # type: Optional[str]

    # task attributes
    target_date = attr.ib()  # type: Optional[date]
    version = attr.ib()  # type: Optional[str]

    driver_name = attr.ib()  # type: str
    is_archived = attr.ib()  # type: bool
    env_name = attr.ib()  # type: Optional[str]
    cloud_type = attr.ib()  # type: str
    trigger = attr.ib()  # type: str

    # runs from same run group will have same root_run_uid
    root_run = attr.ib()  # type: RootRunInfo
    scheduled_run = attr.ib()  # type: Optional[ScheduledRunInfo]

    sends_heartbeat = attr.ib()  # type: bool
    task_executor = attr.ib(default=None)  # type: str
    project_name = attr.ib(default=None)  # type: Optional[str]


@attr.s
class ScheduledRunInfo(_DbndDataClass):
    scheduled_job_uid = attr.ib(default=None)  # type: Optional[UUID]
    scheduled_date = attr.ib(default=None)  # type: datetime
    scheduled_job_dag_run_id = attr.ib(default=None)  # type: int
    # for manual association with scheduled job from dbnd run cli
    scheduled_job_name = attr.ib(default=None)  # type: Optional[str]
    scheduled_job_extra_args = attr.ib(default=None)  # type: Optional[str]

    @classmethod
    def from_env(cls, run_uid):
        scheduled_job_uid = os.environ.get(SCHEDULED_JOB_UID_ENV, None)
        if not scheduled_job_uid:
            # there is no scheduled job
            return
        scheduled_date = os.environ.get(SCHEDULED_DATE_ENV, None)
        if scheduled_date:
            scheduled_date = timezone.parse(scheduled_date)
        scheduled_date = scheduled_date
        scheduled_job_dag_run_id = os.environ.get(SCHEDULED_DAG_RUN_ID_ENV, None)
        return cls(
            scheduled_date=scheduled_date,
            scheduled_job_dag_run_id=scheduled_job_dag_run_id,
            scheduled_job_uid=scheduled_job_uid,
        )


@attr.s
class RootRunInfo(_DbndDataClass):
    root_run_uid = attr.ib(default=None)  # type: UUID

    root_run_url = attr.ib(default=None)  # type: Optional[str]
    root_task_run_uid = attr.ib(default=None)  # type: Optional[UUID]
    root_task_run_attempt_uid = attr.ib(default=None)  # type: Optional[UUID]

    @classmethod
    def from_env(cls, current_run):
        # type: (DatabandRun) -> RootRunInfo
        parent_run = try_get_databand_run()
        if parent_run:
            # take from parent
            root_run_info = parent_run.root_run_info
            # update parent run info if required
            root_task_run = try_get_current_task_run()
            if root_task_run:
                root_run_info = attr.evolve(
                    root_run_info,
                    root_task_run_uid=root_task_run.task_run_uid,
                    root_task_run_attempt_uid=root_task_run.task_run_attempt_uid,
                )

            return root_run_info

        # take from env
        root_run_uid = get_from_env_or_spark_env(DBND_ROOT_RUN_UID)
        root_run_url = get_from_env_or_spark_env(DBND_ROOT_RUN_TRACKER_URL)
        root_task_run_uid = get_from_env_or_spark_env(DBND_PARENT_TASK_RUN_UID)
        root_task_run_attempt_uid = get_from_env_or_spark_env(
            DBND_PARENT_TASK_RUN_ATTEMPT_UID
        )

        if not root_run_uid:
            # current run is the main run
            root_run_uid = current_run.run_uid
            root_run_url = current_run.tracker.run_url

        return cls(
            root_run_uid=root_run_uid,
            root_run_url=root_run_url,
            root_task_run_uid=root_task_run_uid,
            root_task_run_attempt_uid=root_task_run_attempt_uid,
        )


def get_from_env_or_spark_env(key):
    value = os.environ.get(key)
    if value:
        return value

    return get_value_from_spark_env(key)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.constants import RunState
from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._core.utils.dotdict import _as_dotted_dict
from dbnd._vendor.marshmallow import fields, post_load
from dbnd._vendor.marshmallow_enum import EnumField


class ScheduledRunInfoSchema(ApiStrictSchema):
    scheduled_job_uid = fields.UUID(allow_none=True)
    scheduled_date = fields.DateTime(allow_none=True)
    scheduled_job_dag_run_id = fields.String(allow_none=True)
    scheduled_job_name = fields.String(allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return _as_dotted_dict(**data)


class RootRunInfoSchema(ApiStrictSchema):
    root_run_uid = fields.UUID()
    root_task_run_uid = fields.UUID(allow_none=True)
    root_task_run_attempt_uid = fields.UUID(allow_none=True)
    root_run_url = fields.String(allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return _as_dotted_dict(**data)


class RunInfoSchema(ApiStrictSchema):
    root_run_uid = fields.UUID()
    run_uid = fields.UUID()

    job_name = fields.String()
    project_name = fields.String(allow_none=True)
    user = fields.String()

    name = fields.String()
    description = fields.String(allow_none=True)

    state = EnumField(RunState)
    start_time = fields.DateTime()
    end_time = fields.DateTime(allow_none=True)

    # deprecate
    dag_id = fields.String()
    cmd_name = fields.String(allow_none=True)

    execution_date = fields.DateTime()

    # move to task
    target_date = fields.Date(allow_none=True)
    version = fields.String(allow_none=True)

    driver_name = fields.String()
    is_archived = fields.Boolean()
    env_name = fields.String(allow_none=True)
    cloud_type = fields.String()
    trigger = fields.String()
    task_executor = fields.String(allow_none=True)

    root_run = fields.Nested(RootRunInfoSchema)
    scheduled_run = fields.Nested(ScheduledRunInfoSchema, allow_none=True)

    sends_heartbeat = fields.Boolean(default=False, allow_none=True)

    scheduled_job_name = fields.String(allow_none=True)
    scheduled_date = fields.DateTime(allow_none=True)

    external_links = fields.Dict(allow_none=True)

    @post_load
    def make_run_info(self, data, **kwargs):
        return _as_dotted_dict(**data)



# © Copyright Databand.ai, an IBM Company 2022

from typing import Dict, List, Optional, Tuple, Union

import attr

from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._vendor._marshmallow.decorators import pre_load
from dbnd._vendor.marshmallow import fields, post_load


@attr.s(frozen=True, auto_attribs=True)
class DataSchemaArgs:
    type: Optional[str] = None
    columns_names: Optional[List[str]] = None
    columns_types: Optional[Dict[str, str]] = None
    shape: Optional[Tuple[Optional[int], Optional[int]]] = attr.ib(
        default=None, converter=lambda field: tuple(field) if field else None
    )
    byte_size: Optional[int] = None

    @shape.validator
    def validate_shape(self, _, val):
        if val and len(val) != 2:
            raise ValueError(
                f"shape should be a tuple of 2 Integers: (records, columns), got {val}"
            )

    def as_dict(self) -> Dict:
        dictted_schema = {
            "type": self.type,
            "columns": self.columns_names,
            "dtypes": self.columns_types,
            "shape": self.shape,
            "size.bytes": self.byte_size,
        }
        return {k: v for k, v in dictted_schema.items() if v is not None}


class StructuredDataSchema(ApiStrictSchema):
    type = fields.String(allow_none=True)
    columns_names = fields.List(
        fields.String(), load_from="columns", dump_to="columns", allow_none=True
    )
    columns_types = fields.Dict(load_from="dtypes", dump_to="dtypes", allow_none=True)
    # Shape: Tuple of ints-> (records, columns) -> fields.Tuple() added in marshmallow 3.0.0.
    shape = fields.List(fields.Integer(allow_none=True), allow_none=True)
    byte_size = fields.Integer(
        load_from="size.bytes", dump_to="size.bytes", allow_none=True
    )

    @pre_load
    def pre_load(self, data: dict) -> dict:
        # Support Snowflake & PostgreSQL data_schemas before dbnd-sdk-0.61.0
        columns_types = data.get("column_types")
        if columns_types:
            data["columns_types"] = columns_types
        return data

    @post_load
    def make_object(self, data) -> DataSchemaArgs:
        return DataSchemaArgs(**data)


def load_data_schema(
    field: Optional[Union[dict, DataSchemaArgs]]
) -> Optional[DataSchemaArgs]:
    if field is None:
        return None

    if isinstance(field, DataSchemaArgs):
        return field
    return StructuredDataSchema().load(field).data



# © Copyright Databand.ai, an IBM Company 2022

from typing import Any, Callable, List, Optional

import attr

from more_itertools import first

from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._vendor.marshmallow import fields, post_dump, post_load


def get_column_stats_by_col_name(
    columns_stats: List["ColumnStatsArgs"], column_name: str
) -> Optional["ColumnStatsArgs"]:
    # Returns column_stats if column_name is found in columns_stats list
    return first(
        filter(lambda col_stats: col_stats.column_name == column_name, columns_stats),
        None,
    )


@attr.s(auto_attribs=True)
class ColumnStatsArgs:
    column_name: str = attr.ib()
    column_type: str = attr.ib()

    records_count: int = attr.ib()
    distinct_count: Optional[int] = attr.ib(default=None)
    null_count: Optional[int] = attr.ib(default=None)

    # Metric for non-numeric column type
    unique_count: Optional[int] = attr.ib(default=None)
    # Most frequent value
    most_freq_value: Optional[Any] = attr.ib(default=None)
    most_freq_value_count: Optional[int] = attr.ib(default=None)

    # Metric for numeric column type
    mean_value: Optional[float] = attr.ib(default=None)
    min_value: Optional[float] = attr.ib(default=None)
    max_value: Optional[float] = attr.ib(default=None)
    std_value: Optional[float] = attr.ib(default=None)
    # Percentiles
    quartile_1: Optional[float] = attr.ib(default=None)
    quartile_2: Optional[float] = attr.ib(default=None)
    quartile_3: Optional[float] = attr.ib(default=None)

    non_null_count: Optional[int] = attr.ib()
    null_percent: Optional[float] = attr.ib()

    @non_null_count.default
    def _non_null_count(self) -> Optional[int]:
        if self.records_count is None or self.null_count is None:
            return None
        return self.records_count - self.null_count

    @null_percent.default
    def _null_percent(self) -> Optional[float]:
        if self.null_count is None or self.records_count == 0:
            return None

        return (self.null_count / self.records_count) * 100

    def dump_to_legacy_stats_dict(self) -> dict:
        # Returns legacy stats dict for backward compatability support with dbnd-tracking
        stats = {
            "type": self.column_type,
            "count": self.records_count,
            "distinct": self.distinct_count,
            "null-count": self.null_count,
            "non-null": self.non_null_count,
            "top": self.most_freq_value,
            "freq": self.most_freq_value_count,
            "unique": self.unique_count,
            "mean": self.mean_value,
            "min": self.min_value,
            "max": self.max_value,
            "std": self.std_value,
            "25%": self.quartile_1,
            "50%": self.quartile_2,
            "75%": self.quartile_3,
        }
        filtered_stats = {k: v for k, v in stats.items() if v is not None}
        return {self.column_name: filtered_stats}

    def get_stats(self, stats_filter: Optional[Callable] = None) -> dict:
        """Get the column's stats filtered by given filter or by a non-null values filter

        Args:
            stats_filter: Callable that takes two arguments: the stat's name and the stat's value.
            The return code determines whether a stat is included (``True``) or dropped (``False``).
            Defaults to non-null values filter.

        Returns:
            dict: Filtered stats dict
        """
        stats = {
            "distinct_count": self.distinct_count,
            "null_count": self.null_count,
            "unique_count": self.unique_count,
            "most_freq_value": self.most_freq_value,
            "most_freq_value_count": self.most_freq_value_count,
            "mean_value": self.mean_value,
            "min_value": self.min_value,
            "max_value": self.max_value,
            "std_value": self.std_value,
            "quartile_1": self.quartile_1,
            "quartile_2": self.quartile_2,
            "quartile_3": self.quartile_3,
            "non_null_count": self.non_null_count,
            "null_percent": self.null_percent,
        }
        stats_filter = stats_filter or (lambda _, value: value is not None)
        return {k: v for k, v in stats.items() if stats_filter(k, v)}

    def get_numeric_stats(self) -> dict:
        """Get all the column's numeric stats
        Returns:
            dict: Filter stats dict by numeric values
        """

        def _numberic_stats_filter(_, value) -> bool:
            if isinstance(value, (int, float)) and not isinstance(value, bool):
                return True
            return False

        return self.get_stats(stats_filter=_numberic_stats_filter)

    def as_dict(self) -> dict:
        return attr.asdict(self, filter=lambda _, value: value is not None)


class ColumnStatsSchema(ApiStrictSchema):
    column_name = fields.String(required=True)
    column_type = fields.String(required=True)

    records_count = fields.Integer(required=True)
    distinct_count = fields.Integer(required=False, allow_none=True)
    null_count = fields.Integer(required=False, allow_none=True)

    # Metric for non-numeric column type
    unique_count = fields.Integer(required=False, allow_none=True)
    # Most frequent value - load_from attrs to support backward compatibility with SDK < 0.59.0
    most_freq_value = fields.Raw(required=False, allow_none=True, load_from="top_value")
    most_freq_value_count = fields.Integer(
        required=False, allow_none=True, load_from="top_freq_count"
    )

    # Metric for numeric column type
    mean_value = fields.Float(required=False, allow_none=True)
    min_value = fields.Float(required=False, allow_none=True)
    max_value = fields.Float(required=False, allow_none=True)
    std_value = fields.Float(required=False, allow_none=True)
    # Percentiles
    quartile_1 = fields.Float(required=False, allow_none=True)
    quartile_2 = fields.Float(required=False, allow_none=True)
    quartile_3 = fields.Float(required=False, allow_none=True)

    # Hybrid properties - We only dump them so we don't load them back to ColumnStatsArgs when we make_object
    # We don't save them in DB, and they are only computed in ColumnStatsArgs
    non_null_count = fields.Integer(required=False, allow_none=True, dump_only=True)
    null_percent = fields.Float(required=False, allow_none=True, dump_only=True)

    @post_load
    def make_object(self, data: dict) -> ColumnStatsArgs:
        return ColumnStatsArgs(**data)

    @post_dump
    def dump_object(self, data: dict) -> dict:
        return {k: v for k, v in data.items() if v is not None}



# © Copyright Databand.ai, an IBM Company 2022

import attr

from dbnd._core.configuration import environ_config
from dbnd._core.utils.basics.environ_utils import environ_enabled


def disable_airflow_subdag_tracking():
    return environ_enabled(
        environ_config.ENV_DBND__DISABLE_AIRFLOW_SUBDAG_TRACKING, False
    )


@attr.s
class AirflowTaskContext(object):
    dag_id = attr.ib()  # type: str
    execution_date = attr.ib()  # type: str
    task_id = attr.ib()  # type: str
    try_number = attr.ib(default=1)
    context = attr.ib(default=None)  # type: Optional[dict]
    airflow_instance_uid = attr.ib(default=None)  # type: Optional[UUID]

    root_dag_id = attr.ib(init=False, repr=False)  # type: str
    is_subdag = attr.ib(default=None, repr=False)  # type: bool

    def __attrs_post_init__(self):
        if self.is_subdag is None:
            self.is_subdag = (
                "." in self.dag_id
            ) and not disable_airflow_subdag_tracking()
        if self.is_subdag:
            dag_breadcrumb = self.dag_id.split(".", 1)
            self.root_dag_id = dag_breadcrumb[0]
            self.parent_dags = []
            # for subdag "root_dag.subdag1.subdag2.subdag3" it should return
            # root_dag, root_dag.subdag1, root_dag.subdag1.subdag2
            # WITHOUT the dag_id itself!
            cur_name = []
            for name_part in dag_breadcrumb[:-1]:
                cur_name.append(name_part)
                self.parent_dags.append(".".join(cur_name))
        else:
            self.root_dag_id = self.dag_id
            self.parent_dags = []



# © Copyright Databand.ai, an IBM Company 2022

import abc
import datetime

import pytz

from targets.values import DateTimeValueType, DateValueType


_UNIX_EPOCH = datetime.datetime.utcfromtimestamp(0).replace(tzinfo=pytz.utc)


def _add_months(date, months):
    """
    Add ``months`` months to ``date``.

    Unfortunately we can't use timedeltas to add months because timedelta counts in days
    and there's no foolproof way to add N months in days without counting the number of
    days per month.
    """
    year = date.year + (date.month + months - 1) // 12
    month = (date.month + months - 1) % 12 + 1
    return datetime.date(year=year, month=month, day=1)


class _DateCustom(DateValueType):
    """
    Base class ValueType for date (not datetime).
    """

    def __init__(self, interval=1, start=None, **kwargs):
        super(_DateCustom, self).__init__(**kwargs)
        self.interval = interval
        self.start = start if start is not None else _UNIX_EPOCH.date()

    @property
    @abc.abstractmethod
    def date_format(self):
        """
        Override me with a :py:meth:`~datetime.date.strftime` string.
        """

    def to_str(self, dt):
        """
        Converts the date to a string using the :py:attr:`~_DateValueTypeBase.date_format`.
        """
        return dt.strftime(self.date_format)

    def parse_from_str(self, x):
        return datetime.datetime.strptime(x, self.date_format).date()


class MonthValueType(_DateCustom):
    """
    ValueType whose value is a :py:class:`~datetime.date`, specified to the month
    (day of :py:class:`~datetime.date` is "rounded" to first of the month).

    A MonthValueType is a Date string formatted ``YYYY-MM``. For example, ``2013-07`` specifies
    July of 2013.
    """

    date_format = "%Y-%m"

    def _add_months(self, date, months):
        """
        Add ``months`` months to ``date``.

        Unfortunately we can't use timedeltas to add months because timedelta counts in days
        and there's no foolproof way to add N months in days without counting the number of
        days per month.
        """
        year = date.year + (date.month + months - 1) // 12
        month = (date.month + months - 1) % 12 + 1
        return datetime.date(year=year, month=month, day=1)

    def next_in_enumeration(self, value):
        return self._add_months(value, self.interval)

    def normalize(self, value):
        value = super(MonthValueType, self).normalize(value)

        months_since_start = (value.year - self.start.year) * 12 + (
            value.month - self.start.month
        )
        months_since_start -= months_since_start % self.interval

        return self._add_months(self.start, months_since_start)


class YearValueType(_DateCustom):
    """
    ValueType whose value is a :py:class:`~datetime.date`, specified to the year
    (day and month of :py:class:`~datetime.date` is "rounded" to first day of the year).

    A YearValueType is a Date string formatted ``YYYY``.
    """

    date_format = "%Y"

    def next_in_enumeration(self, value):
        return value.replace(year=value.year + self.interval)

    def normalize(self, value):
        value = super(YearValueType, self).normalize(value)

        delta = (value.year - self.start.year) % self.interval
        return datetime.date(year=value.year - delta, month=1, day=1)


class _DatetimeCustom(DateTimeValueType):
    """
    Base class ValueType for datetime
    """

    def __init__(self, interval=1, start=None):
        super(_DatetimeCustom, self).__init__()
        self.interval = interval
        self.start = start if start is not None else _UNIX_EPOCH

    def parse_from_str(self, s):
        """
        Parses a string to a :py:class:`~datetime.datetime`.
        """
        v = datetime.datetime.strptime(s, self.date_format)
        return v.replace(tzinfo=pytz.UTC)

    @property
    @abc.abstractmethod
    def date_format(self):
        """
        Override me with a :py:meth:`~datetime.date.strftime` string.
        """

    @property
    @abc.abstractmethod
    def _timedelta(self):
        """
        How to move one interval of this type forward (i.e. not counting self.interval).
        """

    def to_str(self, dt):
        """
        Converts the date to a string using the :py:attr:`~_DatetimeValueTypeBase.date_format`.
        """
        if dt is None:
            return str(dt)
        return dt.strftime(self.date_format)

    @staticmethod
    def _convert_to_dt(dt):
        if not isinstance(dt, datetime.datetime):
            dt = datetime.datetime.combine(dt, datetime.time.min)
            dt = dt.replace(tzinfo=pytz.utc)
        return dt

    def normalize(self, dt):
        """
        Clamp dt to every Nth :py:attr:`~_DatetimeValueTypeBase.interval` starting at
        :py:attr:`~_DatetimeValueTypeBase.start`.
        """
        dt = super(_DatetimeCustom, self).normalize(dt)

        dt = dt.replace(
            microsecond=0
        )  # remove microseconds, to avoid float rounding issues.
        delta = (dt - self.start).total_seconds()
        granularity = (self._timedelta * self.interval).total_seconds()
        return dt - datetime.timedelta(seconds=delta % granularity)

    def next_in_enumeration(self, value):
        return value + self._timedelta * self.interval


class DateHourValueType(_DatetimeCustom):
    """
    ValueType whose value is a :py:class:`~datetime.datetime` specified to the hour.

    A DateHourValueType is a `ISO 8601 <http://en.wikipedia.org/wiki/ISO_8601>`_ formatted
    date and time specified to the hour. For example, ``2013-07-10T19`` specifies July 10, 2013 at
    19:00.
    """

    date_format = "%Y-%m-%dT%H"  # ISO 8601 is to use 'T'
    _timedelta = datetime.timedelta(hours=1)


class DateMinuteValueType(_DatetimeCustom):
    """
    ValueType whose value is a :py:class:`~datetime.datetime` specified to the minute.

    A DateMinuteValueType is a `ISO 8601 <http://en.wikipedia.org/wiki/ISO_8601>`_ formatted
    date and time specified to the minute. For example, ``2013-07-10T1907`` specifies July 10, 2013 at
    19:07.

    The interval parameter can be used to clamp this parameter to every N minutes, instead of every minute.
    """

    date_format = "%Y-%m-%dT%H%M"
    _timedelta = datetime.timedelta(minutes=1)


class DateSecondValueType(_DatetimeCustom):
    """
    ValueType whose value is a :py:class:`~datetime.datetime` specified to the second.

    A DateSecondValueType is a `ISO 8601 <http://en.wikipedia.org/wiki/ISO_8601>`_ formatted
    date and time specified to the second. For example, ``2013-07-10T190738`` specifies July 10, 2013 at
    19:07:38.

    The interval parameter can be used to clamp this parameter to every N seconds, instead of every second.
    """

    date_format = "%Y-%m-%dT%H%M%S"
    _timedelta = datetime.timedelta(seconds=1)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import urllib

from http import HTTPStatus

from requests import HTTPError, Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from dbnd._core.errors.errors_utils import log_exception


logger = logging.getLogger(__name__)


class DbtCloudApiClient:
    administrative_api_url = "https://cloud.getdbt.com/api/v2/accounts/"
    metadata_api_url = "https://metadata.cloud.getdbt.com/graphql"

    def __init__(self, account_id: int, dbt_cloud_api_token: str, max_retries=3):
        self.account_id = account_id
        self.api_token = dbt_cloud_api_token
        self.session = Session()
        self.session.headers = {"Authorization": f"Token {self.api_token}"}
        retry_strategy = Retry(
            total=max_retries,
            status_forcelist=[429, 500, 502, 503, 504],
            backoff_factor=1,
            method_whitelist=[
                "HEAD",
                "GET",
                "PUT",
                "DELETE",
                "OPTIONS",
                "TRACE",
                "POST",
            ],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)

    def get_account_details(self):
        url = self._build_administrative_url(f"{self.account_id}")
        return self.send_request(endpoint=url)

    def send_request(self, endpoint, method="GET", data={}):
        try:
            if method == "POST":
                res = self.session.post(url=endpoint, json=data)
            elif method == "GET":
                res = self.session.get(url=endpoint, params=data)

            # raise exception for status code != 2**
            res.raise_for_status()
            deserialized_res = res.json()
        except HTTPError as http_error:
            # We want to send server all errors that are not 404 response
            if http_error.response.status_code != HTTPStatus.NOT_FOUND:
                logger.debug("Received unexpected response code from dbt cloud api")
                log_exception("unexpected response code from dbt cloud api", http_error)

            return None

        except Exception as e:
            log_exception("Something went wrong getting data dbt cloud api", e)
            return None

        return deserialized_res

    def _get_run_artifact(self, artifact_name, run_id, step=1):
        path = f"{self.account_id}/runs/{run_id}/artifacts/{artifact_name}"
        url = self._build_administrative_url(path)
        return self.send_request(endpoint=url, data={"step": step})

    def get_manifest_artifact(self, run_id, step=1):
        return self._get_run_artifact(
            artifact_name="manifest.json", run_id=run_id, step=step
        )

    def get_run_results_artifact(self, run_id, step=1):
        return self._get_run_artifact(
            artifact_name="run_results.json", run_id=run_id, step=step
        )

    def get_run(self, run_id):
        if not run_id:
            logger.debug("Can't get run without id")
            return

        path = f"{self.account_id}/runs/{run_id}"
        url = self._build_administrative_url(path)
        res = self.send_request(
            endpoint=url, data={"include_related": '["run_steps", "job"]'}
        )
        return self._safe_get_response_data(res)

    def get_environment(self, env_id: int):
        url = self._build_administrative_url(f"{self.account_id}/environments/{env_id}")
        res = self.send_request(endpoint=url)
        return self._safe_get_response_data(res)

    def list_environments(self):
        url = self._build_administrative_url(f"{self.account_id}/environments/")
        res = self.send_request(endpoint=url)
        return self._safe_get_response_data(res)

    def query_dbt_run_results(self, job_id, run_id):
        query = self._build_graphql_query(
            "models",
            {"runId": run_id, "jobId": job_id},
            ["uniqueId", "executionTime", "status"],
        )
        return self.query_meta_data_api(query)

    def query_dbt_test_results(self, job_id, run_id):
        query = self._build_graphql_query(
            "tests", {"runId": run_id, "jobId": job_id}, ["uniqueId", "status"]
        )
        return self.query_meta_data_api(query)

    def list_runs(self, limit, offset, order_by, job_id=None):
        path = f"{self.account_id}/runs"
        url = self._build_administrative_url(path)
        data = {"limit": limit, "offset": offset, "order_by": order_by}
        if job_id:
            data["job_definition_id"] = job_id
        res = self.send_request(endpoint=url, data=data)
        safe_response = self._safe_get_response_data(res)
        return safe_response or []

    def query_meta_data_api(self, query):
        res = self.send_request(
            endpoint=self.metadata_api_url, method="POST", data={"query": query}
        )
        return self._safe_get_response_data(res)

    def _build_graphql_query(self, resource, filters={}, requested_fields=[]):
        normalized_requested_fields = ",\n".join(requested_fields)
        normalized_filters = ",".join([f"{k}: {v}" for k, v in filters.items()])
        return f"""{{
{resource}({normalized_filters}){{
{normalized_requested_fields}
}}
}}"""

    def _build_administrative_url(self, path):
        return urllib.parse.urljoin(self.administrative_api_url, path)

    def _safe_get_response_data(self, res):
        if res and isinstance(res, dict):
            return res.get("data", None)
        return None



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import functools
import json
import operator

from collections import OrderedDict
from decimal import Decimal
from typing import Mapping
from uuid import UUID

import dbnd._vendor.hjson as hjson

from dbnd._core.utils.platform import windows_compatible_mode
from dbnd._core.utils.traversing import DatabandDict
from dbnd._core.utils.type_check_utils import is_instance_by_class_name


class _FrozenOrderedDict(Mapping, DatabandDict):
    """
    .. deprecated:: 0.7.3

    It is an immutable wrapper around ordered dictionaries that implements the complete :py:class:`collections.Mapping`
    interface. It can be used as a drop-in replacement for dictionaries where immutability and ordering are desired.
    """

    def __init__(self, *args, **kwargs):
        self.__dict = OrderedDict(*args, **kwargs)
        self.__hash = None

    def __getitem__(self, key):
        return self.__dict[key]

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __repr__(self):
        return repr(self.__dict)

    def __hash__(self):
        if self.__hash is None:
            hashes = map(hash, self.items())
            self.__hash = functools.reduce(operator.xor, hashes, 0)

        return self.__hash

    def get_wrapped(self):
        return self.__dict

    def copy(self):
        return _FrozenOrderedDict(self.__dict)


def json_default(obj, safe=False):
    from dbnd._core.parameter.parameter_definition import ParameterDefinition

    if isinstance(obj, ParameterDefinition):
        return str(obj)

    if isinstance(obj, _FrozenOrderedDict):
        return obj.get_wrapped()

    from targets import Target

    if isinstance(obj, Target):
        return str(obj)

    if isinstance(obj, datetime.datetime):
        return obj.strftime("%Y-%m-%dT%H:%M:%SZ")
    elif isinstance(obj, datetime.date):
        return obj.strftime("%Y-%m-%d")

    if is_instance_by_class_name(obj, "int32") or is_instance_by_class_name(
        obj, "int64"
    ):
        return str(obj)

    if isinstance(obj, UUID):
        return str(obj)
    if isinstance(obj, Decimal):
        return float(obj)
    if safe:
        return str(obj)

    raise TypeError(repr(obj) + " is not JSON serializable")


def dumps(obj, default=json_default, sort_keys=True, **kwargs):
    return json.dumps(obj, default=default, sort_keys=sort_keys, **kwargs)


def loads(s, **kwargs):
    if windows_compatible_mode:
        if s and s[0] in ["{", "["]:
            try:
                return loads_canonical(s, **kwargs)
            except Exception:
                return hjson.loads(s, **kwargs)
        return s
    return hjson.loads(s, **kwargs)


def loads_canonical(s, **kwargs):
    return json.loads(s, **kwargs)


def dumps_canonical(obj, separators=(",", ":"), sort_keys=True, default=json_default):
    return json.dumps(obj, separators=separators, sort_keys=sort_keys, default=default)


def dumps_safe(
    obj,
    separators=(",", ":"),
    sort_keys=True,
    default=functools.partial(json_default, safe=True),
):
    return dumps_canonical(
        obj=obj, separators=separators, sort_keys=sort_keys, default=default
    )


def convert_to_safe_types(obj):
    return loads(dumps_safe(obj))


def flatten_dict(json_input: dict) -> dict:
    """
    Flatten nested json dict to 1 level dict
    list values will no further drilled down

     Example input: {"attribute":"value","object_array":[{"attribute":1},{"attribute":2}],"some_nested_object":{"nested_attribute":"nested_value"}}
     output: {"attribute":"value","object_array":"[{'attribute': 1}, {'attribute': 2}]","some_nested_object.nested_attribute":"nested_value"}
    """
    json_output = {}

    def flatten(value, name=""):

        # If the Nested key-value
        # pair is of dict type
        if isinstance(value, dict):

            for key, val in value.items():
                flatten(val, f"{name}{key}.")

        # If the Nested key-value
        # pair is of list type, column value = stringify json and no further drill down
        elif isinstance(value, list):
            json_output[name[:-1]] = json.dumps(value)

        else:
            json_output[name[:-1]] = value

    flatten(json_input)
    return json_output



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd import log_metric
from dbnd._core.utils.basics.text_banner import TextBanner


logger = logging.getLogger(__name__)


class DoctorStatusReportBuilder(object):
    def __init__(self, name):
        self.name = name
        msg = "Running %s Status Check:" % name
        logger.info(msg)
        self.text_banner = TextBanner(msg)

    def log(self, key, value):
        log_metric(key, value)

        self.text_banner.column(key, value)

    def add_sub_report(self, sub_report_str):
        self.text_banner.write(sub_report_str)
        self.text_banner.write("\n")

    def get_status_str(self):
        return self.text_banner.get_banner_str()
